"""
stages of data
Data Source part
0) file: csv-like including comments
1) very raw rows: uncsved list of list of strings
2) raw rows: headers removed, missing values as Nones
3) columns: put in column with a view meta infos, numerized numeric values
4) separated output column
Data Tray part
5) columns: missing value replaced somehow, nonimal values murerized somehow,
    both only for input
6) arrays: input, output as 2d-arrays

Copyright (C) 2011 by Lorenz Weizsacker. 
This module is distributed under the GNU General Public License version 3 
(GPLv3), see U{http://www.opensource.org/licenses/gpl-3.0.html}.

"""

import os

from numpy import array, shape, nan

from tb.fileaccess import load_from_shelve, save_in_shelve, get_lines, \
                          homedir, create_dir
from tb.mixedtools import I, indexed, transpose_list, has_unique_value
from tb.collections2 import BetterList
from tb.baseclass  import Named

from copy import deepcopy
from tb.scipytools import nan_mean





class Column(Named):
    
    def init(self,
             has_missing_values = False,
             name = None,
             entries = [],
             is_numeric = False,
             width = 0,
             ):
        pass



class DataSource(Named):
    """DataSource class. Reads Data from file and makes a
    structured object from it which is stored in place where
    the original file is.
    Only does undisputable preprocessing,"""   
    
    
    global_datadir = None
    global_build   = True
    global_store   = True
    
    
    def init(self, 
             dir    = None,
             source   = None, # relative to dir
             build    = None,
             store    = None,
             data_dir = None,
            ):
        
        if dir is None:
            dir = self.get_name()
    
        
        if source is None:
            source = self.get_name() + '.data'
        
        if not data_dir is None:
            global_datadir = data_dir
        elif not DataSource.global_datadir is None:
            global_datadir = DataSource.global_datadir
        else:
            self.bugger("Please specify data_dir or " +\
                        "DataSource.global_datadir")
        
        self.dir_abs    = os.path.join(global_datadir,dir)
        self.source_abs = os.path.join(self.dir_abs, source)
        
        build   = DataSource.global_build if build is None else build        
               
        store   = DataSource.global_store if store is None else store
         
        if build:
            self.build_data()
            if store:
                self.save()
        else:
            self.load()
 
    def get_name(self):

        return super(DataSource, self).get_name().lower() 
 
    def save(self):
        create_dir(self.dir_abs)
        path = self.store_abs()
        self.send('save build data in', path)
        save_in_shelve(self,path)

    def load(self):
        """returns entire set"""
        path = self.store_abs()
        self.send('get build data set from', path)
        o = load_from_shelve(path) 
        self.set_attributes(o.__dict__) 
        return self

    def store_abs(self):
        return os.path.join( self.dir_abs, self.get_name() + '.pyd')

    def positive(self,output):
        #print output, self.output_values[0]
        return output == self.output_values[-1]


class TableSource(DataSource):
    """Data source where source file contains a table that has one
    row for each example."""
    

    def init(self,
             raw_rows = None,
             rows     = [],
             columns  = [],
             
             nominal_columns   = [], ## indecies of explicit nominal columns 
             output_index      = -1,
             binary_output     = True,
             separator         = None,                
             has_column_header = False,
             has_row_header    = False,
             basedir           = None, ## if set, it is used as path prefix
             missing_value     = [], ## list of string, like '?'    
             
             comment_starts =  ['%', '@', '//'],            
                ):
        
        self.send('n_rows, n_columns:', len(self.rows), len(self.columns))
        pass
    
    def build_data(self):

        self.file_to_rows()
        self.rows_to_columns()
        self.separate_output()
        self.send('build n-rows, n-input-colummns:',
                  len(self.rows), len(self.columns))
       
                
    def file_to_rows(self):
        
        self.send('build data set from', self.source_abs)
                
        lines         = self.get_data_lines(self.source_abs)
        
        splitted_lines = self.get_splitted_lines(lines)
        
        self.very_raw_rows = deepcopy(splitted_lines)
        
        if self.has_column_header:
            self.column_header = splitted_lines.pop(0)
            
        if self.has_row_header:
            self.row_header = []
            for line in splitted_lines:
                self.row_header += [ line.pop(0) ]
                
        for i,line in indexed(splitted_lines):        
            for j,entry in indexed(line):
                entry = line[j] = entry.strip()
                if entry in self.missing_value:
                    line[j] = None
            self.rows += [ line ]
        
        
        self.raw_rows = deepcopy(self.rows)
        
        self.n_rows = len(self.rows)
        
    def rows_to_columns(self):
        
        if self.rows:
            a_row = self.rows[0]
            for j,entry in indexed(a_row):
                column = Column()
                #column.index = j
                if self.has_column_header:
                    column.name = self.column_header[j]
                else:
                    column.name = None
                self.columns += [ column ]

        columns_plain     = transpose_list(self.rows)
        
        for j,column in indexed(self.columns):
            
            column.entries = entries = columns_plain[j] 
            
            column.width = 0
            
            for entry in entries:    
                if entry == None:
                    column.has_missing_values = True
                column.width = max(column.width, len(str(entry)))
                
            if j in self.nominal_columns:
                column.is_numeric = False 
            else:
                column.is_numeric = self.is_numeric(entries)
            
            if column.is_numeric:
                for i,entry in indexed(entries):                    
                    entries[i] = self.numerize(entry)

     
     
    def separate_output(self):
        self.output_column = self.columns.pop(self.output_index)               
    
    def output_mapping(self,index):
        return self.output_column.entries[index]
    
    def is_numeric(self,column_entries):
        
        out = True
        for entry in column_entries:
            if not entry == None:
                entry_ = self.numerize(entry)
                if entry_ is None:
                    out = False
                    break
        
        return out

            
    def numerize(self,s):
        """tries to convert string s into a number"""
        if s == None:
            return s
        
        ## detect comma numbers
        if s.count(','):
            l,r = s.split(',')
            ln = self.numerize(l)
            rn = self.numerize(r)
            if type(ln) is int and type(rn) is int:
                s = '.'.join([str(ln),str(rn)])
                
        try:
            v = int(s)    
        except:
            try:
                v = float(s)
            except:
                v = None    
        return v
              
    def get_data_lines(self,file_path):
        
        def is_data_line(line):
            
            return len(line) > 0 and\
                   not any([line.strip().startswith(s) 
                            for s in self.comment_starts ])
        
        lines_striped =  [ line.strip() for line in get_lines(file_path)]

        lines = [ line for line in lines_striped if is_data_line(line) ]
        return lines

    def get_splitted_lines(self, lines):
        """splits data lines of input file by separator
        which is automatically detected using an heuristic"""
        
        if self.separator is None:
            self.separator = self.detect_separator(lines)
        else:
            assert type(self.separator) is type('')

        splitted_lines = [ line.split(self.separator)
                           for line in lines ]
        
        return splitted_lines
 
    def detect_separator(self,lines):
        
        
        chr2int = lambda x: [chr(i) for i in range(255)].index(x)
        
        favorate_separators = [chr2int(c) for c in [',', '\t', ';'] ]
        
        indices    = range(255)
        C_rows     = [ [  line.count(chr(i)) for i in indices ]  for line in lines]
        C_columns  = transpose_list(C_rows)  
        
        candidates = [ i for i in indices 
                          if has_unique_value(C_columns[i]) and C_columns[i][0] > 0  ]
        
    
        l = len(candidates)
        if l == 0:
            self.send('cant find seperator.',lb=2)
        if l == 1:
            separator = candidates[0]    
        if l > 1:
            # drop alphanumerics from candidate list
            candidates = [ c for c in candidates if not chr(c).isalnum()]  
            l          = len(candidates)
            if l == 0:
                self.send('there are at least two alphanumeric'+\
                          ' seperator candidates, but no other.')
            if l == 1:
                separator = candidates[0]
            
            if l > 1:
                # more than one remaining candidate: 
                # take the first of most frequent ones. If runner-up is exactly 
                # one count less, take runnner-up
                
                # sort candidates descending according to counts
                
                counts = BetterList([ C_columns[i][0] for i in candidates ])
                counts_sorting    = counts.get_sorting_permutation(reverse=True)
                counts_sorted     = counts.get_sorted_list(reverse=True)
                candidates_sorted = [candidates[i] for i in counts_sorting]
    
                self.send("Separator candidates: ", [chr(i) for i in candidates_sorted], \
                    ", Indices: ", candidates_sorted, 
                    ", Counts: ", counts_sorted)
                    
                # take most frequent candidate or runner up
                maximizer         = candidates_sorted[0]
                maximum           = counts_sorted[0]
                runner_up         = candidates_sorted[1]
                runner_up_count   = counts_sorted[1]
                 
                if maximum == runner_up_count + 1:
                    # some chars in each column, take runner up    
                    separator = runner_up
                else: 
                    # take most frequent one
                    separator = maximizer
                
                # prefer favorate separators regardless their counts
                favorate_separators.reverse()
                for c in favorate_separators:
                    if c in candidates_sorted:
                        separator = c
                        break
                             
        self.send("detected separator: '%s' (%d)" % (chr(separator), separator))             
        return chr(separator)
      
        
    def show_table(self,n=None):
        
        n = self.n_rows if n is None else min(n, self.n_rows)
        
        
        out = ''
                
        for i in range(n):
            for j,column in indexed(self.columns):
                entry = column.entries[i]
                out += ' ' * (column.width - len(str(entry)) + 2) + str(entry)             
            out += '\n'
        
        self.send(out)   
    

 
class FunctionalDataTray(Named):
    """base set is a list of points each of which consist of
    an input and an output part. The joint distribution
    of the points is assumed to be exchangeable such that 
    random sub-sampling makes sense."""
    
    def init(self,
             data_source = None,
             inputs   = [],
             outputs  = [],  
             training_sample = [], ## list of indices for training
             ):
        pass

        
    def n_distinct_outputs(self, selection=None):

        if selection is None:
            selection = I(self.outputs)

        
        output_set = set([ self.output_column.entries[i] 
                           for i in selection ] )
        n_outputs  = len(list(output_set))

        return n_outputs
          
    
    def get_X(self, selection=None):
        
        #if not self.has('X'):
        #    self.build_X_array() 
            
        if selection is None:
            selection = I(self.X)

        return self.X[selection]   

    def get_Y(self, selection=None):
        
        #if not self.has('Y'):
        #    self.build_Y_array() 
            
        if selection is None:
            selection = I(self.Y)
        
        return self.Y[selection]   
    
    def get_XY(self, selection=None):
        
        return self.get_X(selection), self.get_Y(selection)



    def remove_duplicates(self,A, contradictions=False):
        
        X,Y = self.get_XY(A)
        
        D = []
        for i in I(X):
            for j in range(i+1,len(A)):
                if (X[i] == X[j]).all() and ((Y[i] != Y[j]).all() or not contradictions):
                    D += [A[i]]  
        A_  = [ i for i in A if not i in D]
        #quit()
        return A_
        

class TableTray(FunctionalDataTray):
    """Further digests data from a TableSource. It makes more disputable
    pre-processing that partially depends on training_sample. 
    * replace missing numeric values by means on training_sample
    * replace nominals from a set of size k by k disjoint binary values.
        Missing values are represented as k zeros."""
    
    def init(self,
                source  = None,              
                ):
        
        assert source
        
        self.output_index  = source.output_index 
            
        self.binary_output = source.binary_output
        self.n_rows = len(source.columns[0].entries)
        
        self.build_input()
        self.build_output()
        self.build_output_array()
        self.build_input_array()
        
        #self.send('n-input-colummns post-numerization:', len(self.columns),lb=2)
    
    def build_input(self):
        self.replace_missing_numerics()
        self.make_binary_columns_nominal()
        self.numerize_nominals()



    def replace_missing_numerics(self):
        ## replace missing value for numeric columns
        for column in self.source.columns:
            if column.is_numeric:
                for i,entry in indexed(column.entries):
                    if entry == None:
                        column.entries[i] = nan

                if self.training_sample == []:
                    self.training_sample = range(self.source.n_rows) 
                
                training_subset = [ column.entries[i] 
                                    for i in self.training_sample ]
                
                mu = nan_mean(training_subset)
                #mu = nanmean(training_subset)
                for i,entry in indexed(column.entries):
                    if entry is nan:
                        column.entries[i] = mu
                
                        
    def numerize_nominals(self):    
        new_columns = []
        J = 0
        n_rows = self.source.n_rows
        for j,column in indexed(self.source.columns):
            if column.is_numeric:
                new_columns += [ column ]
                J += 1
            else:
            
                values  = set(column.entries)
                if column.has_missing_values:
                    values.remove(None)
                values  = sorted(values)
    
                for k in I(values):
                    new_column = Column()
                    new_column.entries = [ 0 ] * n_rows
                    for i in range(n_rows):
                        ## no ones for missing values
                        new_value = ( values[k] == column.entries[i] )
                        new_column.entries[i] = int(new_value)
                    new_column.width = 2
                    new_column.is_numeric = True
                    new_columns += [ new_column ]
                    J += 1
        self.columns = new_columns

    def make_binary_columns_nominal(self):
        for j,column in indexed(self.source.columns):
            if column.is_numeric and len(set(column.entries)) <= 2:
                column.is_numeric = False
                    
           
    def build_output(self):
        column = self.source.output_column
        values = sorted(set(column.entries))
        self.source.output_values = values
        
        assert not column.has_missing_values
        if not column.is_numeric:
            
            new_column = Column()
            new_column.entries = []
            for i in range(self.source.n_rows):
                entry_raw = column.entries[i]
                new_column.entries += [ values.index(entry_raw) ]

                new_column.width = max(new_column.width, 
                                       len(str(new_column.entries[-1])))
            self.output_column = new_column

        if self.binary_output:
            new_column = Column()
            new_column.entries = []
            for i in range(self.source.n_rows):
                entry_raw = column.entries[i]
                entry = int(self.source.positive(entry_raw))
                new_column.entries += [ entry ]
               
                new_column.width = max(new_column.width, 
                                       len(str(new_column.entries[-1])))
            self.output_column = new_column

                

    def build_input_array(self):
        columns_plain = [ c.entries for c in self.columns ]
        rows = transpose_list(columns_plain)
        self.inputs = rows
        self.X = array(rows,dtype='d')
    
    def build_output_array(self):
        column_plain =  self.output_column.entries 
        values = sorted(set(column_plain))
        for i,entry in indexed(column_plain):
            if self.binary_output:
                if entry:
                    column_plain[i] = +1
                else:
                    column_plain[i] = -1
        self.outputs  = column_plain
        column_plain_ = [ column_plain ]
        #rows = transpose_list(column_plain_)
        self.Y = array(self.outputs,dtype='d')


    def show_table(self, n=None):
        
        n = self.n_rows if n is None else min(n, self.n_rows)
        
        out = ''        
        for i in range(n):
            for j,column in indexed(self.columns):
                entry = column.entries[i]
                out += ' ' * (column.width - len(str(entry)) + 2) + str(entry)             
            out += '\n'
        self.send(out)      

    def show_up(self,n=None):
        
        def entry_string(column,i):
            entry = column.entries[i]
            return ' ' * (column.width - len(str(entry)) + 2) + str(entry)                    
        
        n = self.n_rows if n is None else n
        
        s = range(self.n_rows)
        random.shuffle(s)
        row_selection = s[:n]
        
        
        raw_rows      = [ self.source.raw_rows[i] for i in row_selection ]

        
        very_raw_rows = [ self.source.very_raw_rows[i] for i in row_selection ]

        for column in self.columns:
            column.width = max([ len(str(e)) for e in column.entries] )
            
        
        self.send('\ntable rows as read from file:')
        for row in very_raw_rows: 
            self.send(row)
            
            
        self.send('\n\nrows with nonized missing values and without headers:')
        for row in raw_rows: 
            self.send(row)
            
        self.send('\n\ninput columns with numerized num-like entries:')
        out = ''        
        for i in row_selection:
            for j,column in indexed(self.source.columns):
                out += entry_string(column, i)
            out += '\n'
        self.send(out)      


        self.send('\nfinal numeric table:')
        out = ''        
        for i in row_selection:
            for j,column in indexed(self.columns):
                out += entry_string(column, i)
            out += '\n'
        self.send(out,lb=2)  
        self.send('n rows' , self.n_rows)
        self.send('n column', len(self.columns))
        self.send('output values', self.source.output_values)
        self.send('shapes X,Y',shape(self.X),shape(self.Y))
        self.send('Y[subset]:\n',self.Y[row_selection])
        
    
    
class clean1(TableSource):
    def init(self, 
             source = 'clean1.arff',
            ):
        pass
    

class acute(TableSource):
    def init(self, 
             source = 'diagnosis.data',
            ): 
        pass
    

class sonar(TableSource):
    def init(self, 
             source   = 'sonar.arff',
            ):
        pass


class diabetes(TableSource):
    def init(self, 
                source   = 'diabetes.arff',
                ):
        pass
    
    def positive(self,output):
        return output == self.output_values[1]


class heartstatlog(TableSource):
    def init(self, 
                source   = 'heart-statlog.arff',
                ):
        pass
    

class parkinsons(TableSource):
    def init(self, 
             source   = 'parkinsons.data',
             output_index      = -7,
             has_column_header = True,
             has_row_header    = True,
            ):
        pass
 

class vehicle(TableSource):
    def init(self, 
                source   = 'vehicle.arff',
                ):
        pass
    
class ionosphere(TableSource):
    def init(self, 
                source  = 'ionosphere.arff',
                ):
        pass

class yeast(TableSource):
    def init(self, 
                source   = 'yeast.arff',
                ):
        pass

    def positive(self,output):
        #print output, self.output_values
        return output in ['MIT','ME3','ME2','ME1'] #'CYT','NUC','EXC','VAC','POX','ERL']
                
    
class vote(TableSource):
    def init(self, 
                source   = 'vote-1.arff',
                missing_value = ['?']
                ):
        pass


class arrhythmia(TableSource):
    """Has 73 nomimals but we dont which of the 279 they attributes are."""  
    def init(self, 
                source   = 'arrhythmia.data',
                missing_value = ['?']
                ):
        pass

class breast_cancer(TableSource):
    def init(self, 
                source   = 'breast_cancer.arff',
                nominal_columns = [ 5 ]
                ):
        pass


class colic(TableSource):
    def init(self, 
                source   = 'colic.arff',
                missing_value = ['?']
                #nominal_columns = [ 5 ]
                ):
        pass
 
 

            
if __name__ == '__main__':
    
    from ml.artificial_data import *
    DataSource.global_datadir = os.path.join(homedir(), 'code/data_sets_1/')
    
    #ds = clean1()
    #ds = parkinsons()
    #ds = vehicle()
    #ds = ionosphere()
    ds = heartstatlog()
    #ds = yeast()
    #ds = vote()
    #ds = sonar()
    #ds = RD2D_A(data_dir=global_datadir)
    #ds = acute()
    #ds = breast_cancer()
    #ds = colic()
    #ds = diabetes()
    
    dt = TableTray(source=ds)
    
    dt.show_up(10)
    
    
    
    
