"""A libsvm wrapper that uses a hacked libsvm to access model alpha. """

import os
from tb.baseclass import Named
from tb.scipytools import col, indices_equal, numberize, tp, a2sign, row
from tb.mixedtools import timestamp, I
from tb.fileaccess import get_lines
from numpy import sum,sqrt, shape, array, zeros, dot, mean, diag, repeat
from ml.model import LLMM
 
import svm as libsvm

from numpy.core.numeric import concatenate

###4
class LIBSVM_BB(Named): 
    """libsvm-wrapper that yields beta,b, based on a hacked version 
    of libsvm."""
    

    def init(self,
                data_tray   = None,
                kernel_tray = None,
                C           = 1.,
                nu          = 0.1,
                svm_type    = 'C',
                balancing_weights = False,
                ):
        
        
        self.engine = libsvm
        
        libsvm.svm_parameter(kernel_type = self.engine.LINEAR)
        
        parameter   = self.engine.svm_parameter(kernel_type = self.engine.LINEAR)
            
        if   svm_type == 'C':
            parameter.svm_type = self.engine.C_SVC 
            parameter.C = 1. * C
        elif svm_type == 'nu':
            parameter.svm_type = self.engine.NU_SVC
            parameter.nu = 1. * nu
            self.send("nu: ", nu)
        else:
            self.bugger('Wrong svm_type.')
        
        parameter.kernel_type = self.engine.PRECOMPUTED

        self.parameter = parameter
       
    def set_C(self, C, n):
        self.C = C
        self.parameter.C = C/float(n) 
        
         
    def train(self, S, sign=+1):
        
            
        
        kt = self.kernel_tray
        
        M = LLMM(sign=sign, maker=self)
                
        assert sign in [ -1, +1]
        
        S_       = self.group_by_label(S)
        X,Y      = self.data_tray.get_XY(S_)
        Y        = sign * Y 
        n,d      = shape(X)

        if self.balancing_weights:
            no = len(indices_equal(Y, -1))
            po = len(indices_equal(Y, +1))
            oo = len(Y) * 1.
            self.parameter.nr_weight = 2
            self.parameter.weight_label = [-1,+1]
            #we = [po,no]
            #print [type(e) for e in we]
            self.parameter.weight = [po/oo,no/oo]
            
        M.S = S_
            
        M.sigma = kt.pp.sigma
        M.mu    = kt.pp.mu
            
        #self.clock('query')
        
        K = M.K = kt.query_cross(S_,S_)
        
        #self.send_clock('query')
        
        K_ = concatenate((col(range(1,n+1)),K),axis=1)
        #print K_
        
        #self.clock('train')
        
        problem         = self.engine.svm_problem(Y, K_)
       
        M.libsvm_model  = self.engine.svm_model(problem, self.parameter)
       
        M.beta, M.tau     = self.get_betab(M.libsvm_model,n,d)
        
        #self.send_clock('train')
        
        return M
        
        
    
    def predict(self, model, T, 
                      get_scores=False,
                      more = False):
        
        beta, tau, S = model.beta, model.tau, model.S
        
        kt = self.kernel_tray
        
        kt.pp.mu    = model.mu
        kt.pp.sigma = model.sigma
        
        K_cross = kt.query_cross(S,T)
        
        g       = dot(tp(beta), K_cross).ravel()
        f       = g - tau
        
        
        Y = self.data_tray.get_Y(T)
        
        m = Y.ravel() * f
        Y_hat   = a2sign(f)
        
        if more:
            return Y_hat,m
        elif get_scores:
            return Y_hat, f
        else:
            return Y_hat



    
    def predict_black_box(self, model, T, 
                        get_scores=False, get_scale=False):
        
        libsvm_model  = model.libsvm_model
        S             = model.S

        X_test = self.data_tray.get_X(T)
        
        X_test = model.apply_scale(X_test)
        
        Y_hat  = col(array( [ libsvm_model.predict(x) for x in X_test]  ))
        
        if get_scores or get_scale:
            #print "TTT", libsvm_model.predict_values(x).keys()
            f_test = col(array( [ libsvm_model.predict_values(x)[(+1,-1)] for x in X_test] ))
            
            if get_scale:
                scale  = 1./self.get_gamma_from_libsvm_model(model) #=||w||
                return Y_hat, f_test, scale
            else:
                return Y_hat, f_test
        else:
            return Y_hat





    def get_betab(self, libsvm_model, n, d, get_header=False):
        """gets the beta vector from a lib-svm model object"""
        
        header, svs, b  = self.read_model(libsvm_model,d) 
        beta            = col(zeros(n))
        for sv in svs:
            beta[int(sv['index'])] =  sv['beta'] 
              
        if get_header:
            return beta, b, header
        else:
            return beta, b

 
    def get_gamma(self, model):

        beta,  K = model.beta, model.K
        scale      = sqrt(dot( dot( tp(beta), K ), beta )).ravel()[0] 
        gamma      = 1./scale
        
        return gamma
               


    def get_data_scale(self,S):   
        
        if self.data_scale == None:
            if self.name == 'rbf':
                d_avg = 1.
            else:
                d     = self.get_distance_from_center(S)
                d_avg = mean(d).ravel()[0]
            self.data_scale = d_avg
                
        return self.data_scale
    
  
     
    def get_distance_from_center(self, S):
        
        X = self.data_tray.get_X(S)
        K = self.kt.km(self.kernel.apply, X, X)
        n = len(S)
        D = [ K[i,i] -  2./n * sum(K[i,:]) + 1./n**2 * sum(sum(K)) for i in range(n) ]
        d = col(sqrt(D))
        
        
        return d

    def get_model_distances(self,E):
        pass
        
        
    def get_point_distances(self, A,E):
        
        X_A = self.data_tray.get_X(A)
        X_E = self.data_tray.get_X(E)
        n_A = len(X_A)
        n_E = len(X_E)
        
        K_A  = self.kt.km(self.kernel.apply, X_A, X_A)
        K_E  = self.kt.km(self.kernel.apply, X_E, X_E)
        K_AE = self.kt.km(self.kernel.apply, X_A, X_E)
        k_A = diag(K_A)
        k_E = diag(K_E)
        
        #distance from the mean
        DC_A     = [ K_A[i,i] -  2./n_A * sum(K_A[i,:]) + 1./n_A**2 * sum(sum(K_A)) for i in range(n_A) ]
        dc_A     = col(sqrt(DC_A))
        dc_A_avg = mean(dc_A).ravel()[0]
        
        #test distances
        D  = sqrt( repeat(col(k_A), n_E, axis=1) + repeat(row(k_E), n_A, axis=0) - 2 * K_AE )
        
        out = tp(D)
        
        assert shape(out) == (n_E,n_A)
        
        return out, dc_A_avg
        
        


      

    def read_model(self, model, d):
        
        def get_all_lines(model):   
            model_path = '/dev/shm/' + 'libsvm_model_' + timestamp() + '__' + str(os.getpid()) + '.txt'    
            model.save(model_path)
            out = get_lines(model_path)
            os.remove(model_path)
            return out
      
      
        def separate_parts(lines):
            
            header_lines = []
            data_lines   = []
        
            passed = False
            
            for line in lines:
                
                if line[:2] == 'SV': 
                    passed = True
                else:
                    if passed:
                        data_lines   += [ line ]
                    else:
                        header_lines += [ line ] 
            
            return header_lines, data_lines         

        def header2dict(header_lines):
            out = {}
            for line in header_lines:
                parts = line.split()
                
                key    =  parts[0]
                values =  parts[1:]
                for i in I(values):
                    values[i] = numberize(values[i])
                out[key] = values
                
            return out
             
        def support_vectors(data_lines, model_sign, d):
            #print "MDS", model_sign
            svs = []
            

            for line in data_lines:
                #print line
                sv         = {}
                parts       = line.split()
                sv['index'] = parts[0]
                beta       = float(parts[2])
                sv['beta'] = model_sign * beta
                #vd         = {}
                #for part in parts[3:]:
                #    index,value = part.split(':')
                #    vd[index]   = value
                        
                #vector = zeros(d)
    
                #for key,value in vd.items():
                #    vector[int(key)] = value
                        
                #sv['vector'] = array(vector)
                    
                svs += [ sv ]
            
            #bugger
            return svs
        
        lines = get_all_lines(model)
        
        header_lines, data_lines =  separate_parts(lines)
        
        header  =  header2dict(header_lines)
        labels  = header['label']
        
        try:
            assert set(labels) == set([-1,1])
        except:
            print header
        model_sign = labels[0] # what class does model regard as positive
                              
        svs  =  support_vectors(data_lines, model_sign,d )
        
        # libsvm defined first label as postive 
        b = model_sign  * header['rho'][0]
        
        return header, svs, b


    def group_by_label(self,S):
        Y         = self.data_tray.get_Y(S)    
        positives = [ S[i] for i in indices_equal(Y,+1) ]
        negatives = [ S[i] for i in indices_equal(Y,-1) ]
        return positives + negatives  



if __name__ == '__main__':
    
    pass    
