from tb.mixedtools import I, indexed
from tb.baseclass  import Named 

import random
from copy import deepcopy

class SubSampling(Named):
    """Class for subsampling from input sequence. 
    Allows stratified sampling with respect to 
    a mapping on ground_set. 
    
    Though subset selection can be randomized, the original order can 
    be kept in the subsets.
    """
    
    def init(self,
                sequence    = None,
                restriction = [], ## list of indices on sequence
                mapping     = lambda index: "soso",
                
                randomize  = True,
                keep_order = False,
                seed       = None,
                n_folds    = 8,
                max_train_size = 2**22,
                max_test_size  = 2**22,                
                ):

        assert sequence
        
        self.set_sequence(sequence) 
        self.build_equivalence_classes()
        
                
    def set_sequence(self, sequence):
          
        self.sequence        = sequence 
        self.ordered_indices = I(self.sequence)
        self.all_indices     = self.ordered_indices[:]
        

        if not self.restriction:
            self.restriction = self.all_indices
        
        self.relevant_indices = [ i for i in self.all_indices
                                  if i in self.restriction    ]

        if self.randomize:
            random.seed(self.seed)
            random.shuffle(self.relevant_indices)

        
    
    def get_sequence_from_subset(self, subset):
        """Return items refered by the indices in subset"""
        if self.keep_order:
            subset = self.restore_order(subsets)       
        return [ self.sequence[i] for i in subset ]


    def restore_order(self, subset):
        """restore original order in subset"""
        return [ i for i in self.ordered_indices if i in subset ]


    def build_equivalence_classes(self):
        """Build index subsets one for each output of self.mapping"""
        
        codomain  = list( set([ self.mapping(i) 
                                for i in self.relevant_indices ] ) )

        classes   = [ [] for y in codomain ]
        
        for i in self.relevant_indices:
            classes[ codomain.index(self.mapping(i)) ] += [ i ]
        
        self.classes = classes

    


    def stratified_partition(self, k):
        """Computes weakly stratified partition with k subset.
        This is done by round robin item distribution such that the 
        number of items from given class differ by at most 1."""
        
        partition = [ [] for i in range(k) ] 
       
        
        # distribute indices
        for c in deepcopy(self.classes):
            while len(c) > 0:
                taken = c.pop()
                subset = partition.pop(0)
                subset += [taken ]
                partition.append(subset)
            
        return partition

 
 
    def stratified_subset(self, setsize):
        K = len(self.classes)
        n = len(self.relevant_indices)        
        classes = deepcopy(self.classes)
        out = [] 
        class_sizes_rel = [ len(c)*1./n for c in self.classes] 
        class_sizes_out = [ int(sr * setsize) for sr in class_sizes_rel ]
        
        res        = setsize - sum(class_sizes_out)
        
        assert res <= K
        
        for j in range(res):
            class_sizes_out[j] = class_sizes_out[j] + 1
        
        for i in range(K):
            out += classes[i][:class_sizes_out[i]]
        
        assert abs(len(out) - setsize) <= 1, [len(out), setsize]
            
        return out

 
  
    def get_cv_indices(self):
        """Computes cross stratified validation folds as list of
        paits of list of inidices. 
        If the sizes are bound, the subsets are not stratified anymore.""" 
        
        k = self.n_folds
        
 
        subsets = self.stratified_partition(k)
        
        folds = []

        for test_part in subsets:
            
            train_part = list(set(self.relevant_indices).difference(set(test_part)))

            if self.max_train_size:
                train_part = train_part[:self.max_train_size]
            if self.max_test_size:
                test_part = test_part[:self.max_test_size]
                
            
            folds += [(train_part,test_part)]
            
        return folds
        

    def get_cv_folds(self):
        """Computes cross validation folds. If the sizes are bound, 
        the subsets are not stratified anymore.""" 

        indices = self.get_cv_indices()                    
        
        out =  [ ( self.get_sequence_from_subset(A), 
                   self.get_sequence_from_subset(B)  ) 
                   for (A,B) in indices ]
        
        return out
        

## unused

    def get_subset_pair(self, max_train_size  = None,
                              max_test_size   = None):
        """unstratified, disjoint pair of subsets"""
        
        n = len(self.relevant_indices)
        if not max_train_size and not max_test_size:
            max_test_size  = int(n/2.)
            max_train_size = n - max_test_size
        elif not max_train_size:
            max_train_size = n - max_test_size
        elif not max_test_size:
            max_test_size = n - max_train_size
        else:
            assert  max_test_size + max_test_size <= n
        

        test_part  = self.relevant_indices[:max_test_size] 
        train_part = list(set(self.relevant_indices).difference(test_part))
        train_part = random.shuffle(train_part)[:max_train_size] 
        
        return ( self.get_sequence_from_subset(train_part),
                 self.get_sequence_from_subset(test_part) )
        
        
    def get_doubled_subset(self, max_size=None):
        """unstratified, identical pair of subsets"""
        if max_size == None:
            n = len(self.relevant_indices)
            max_size  = int(n/2.)
        
        test_part  = self.relevant_indices[:max_size]
        train_part = test_part[:]
     
        return ( self.get_sequence_from_subset(train_part),
                 self.get_sequence_from_subset(test_part) )
        


if __name__ =='__main__':
    
    def mapp(x):

        return  int(x) % 3 == 0
        
     
    
    S = SubSampling( sequence=range(50,70), mapping= mapp )
    
    #subsets = S.get_doubled_subset(20)
    subsets = S.get_subset_pair(20,10)
    folds = S.get_cv_folds(3, 10, 20)
    
    for subsets in folds:
        print "BUU"
        for s in subsets:
            print s
    
    




