"""The module define some classes that are used as base souce."""   

from copy import deepcopy
from subprocess import Popen, PIPE
import sys, time
from mixedtools import seconds2time

class LeanDescendant(object):
    """The very ancestor for many class we use. Its main feature is the 
    initialization through the init method, which makes descendants to inherit
    default values if they dont overwrite them and also automatically calls ancestor 
    init functions that do not have mandatory arguments."""


    def __init__(self, *mandatory_args, **optional_args):
        """The initialization is delegated to init. 
        For any ancestor, the optional arguments of ancestor.init are used to set 
        corresponding attributes and ancestor.init is called, if it take no mandatory 
        arguments. The proper init is called in any case and mandatory args 
        become attributes likewise the optional arguments.
        
        The default value for optional arguments are deepcopied such that each instance 
        has its proper default values. Unfortunately, deepcopying does not work for 
        any type, so take care that default values are deepcopyable.
        """
    
        ancestors = self.get_ancestors()

        ## get all default_dicts and merge them
        merged_default_dict = {}
        for ancestor in ancestors + [ self.__class__ ]:
            
            default_dict = self.get_args(ancestor.init)[2]
            merged_default_dict.update(default_dict.items())
        
        
        for key in optional_args:
            if not key in merged_default_dict:
                print(ancestors)
                print(merged_default_dict)
                error_text = 'Unknown optional argument for %s: %s' % (self.__class__, key)
                raise Exception(error_text) 
                 
        ### call ancestor inits   
        for ancestor in ancestors:
            
            mandatory_names = self.get_args(ancestor.init)[0]  
            optional_names  = self.get_args(ancestor.init)[1]  
            optionals_dict  = self.merge_dicts(merged_default_dict, optional_args )                 
            self.set_attributes(optionals_dict)

            ## call ancestor.init, only if it takes no mandatory args
            if len(mandatory_names) == 0:
                optionals_dict_ = {}
                for key in optionals_dict:
                    if key in optional_names:
                        optionals_dict_[key] = optionals_dict[key]
                        
                ancestor.init(self, **optionals_dict_)
                
                for key in merged_default_dict:
                      
                    merged_default_dict[key] = self.__dict__[key]
        
        
        ### proper init function call        
        mandatory_names = self.get_args(self.init)[0]  
        optional_names  = self.get_args(self.init)[1]  
        optionals_dict  = self.merge_dicts(merged_default_dict, optional_args )
        self.set_attributes(optionals_dict)
        
        ## set mandatory arguments as members 
        assert len(mandatory_args) == len(mandatory_names), '\n' +\
            'Wrong number of mandatory argument for %s.init()\n' % self.__class__.__name__ + \
            'Expected mandatory arguments: %s\n' % str(mandatory_names) +\
            'But %d values were passed.\n' % len(mandatory_args)
            
        for i in range(len(mandatory_names)):
            self.__setattr__(mandatory_names[i], mandatory_args[i])
        
        ## call self.init with mandatory args
        
        optionals_dict_ = {}
        for key in optionals_dict:
            if key in optional_names:
                optionals_dict_[key] = optionals_dict[key]
        self.init(*mandatory_args, **optionals_dict_)
        
    
    def merge_dicts(self, defaults, provided_dict):
        """Return dict with keys from defaults. The values are taken
        from all_defaults if key exist and from defaults otherwise"""
        out = {}
        
        for key in defaults.keys():
            
            if provided_dict.has_key(key):
                out[key] = provided_dict[key]
            else:
                out[key] = defaults[key]
                
        return out

                
    def get_args(self, f):
        """Assumes that f is instance method and returns list of mandatory args,
        excluding 'self', and dict that maps optional args to default values."""         
        
        all_varnames    = f.func_code.co_varnames
        all_arguments   = all_varnames[:f.func_code.co_argcount]
        defaults        = f.func_defaults
        n_optional      = len(defaults) if defaults else 0 
        n_mandatory     = len(all_arguments) - n_optional   # including 'self'
        mandatory_names = all_arguments[1:n_mandatory]      # exclude 'self'
        optionals_names = all_arguments[n_mandatory:]
        
        
        if defaults:
            defaults = deepcopy(defaults)
            default_dict = dict( zip( optionals_names, defaults))
        else:
            default_dict = {}
        
        
        #print self, mandatory_names, defaults, default_dict
        return mandatory_names, optionals_names, default_dict

    def get_ancestors(self):
        """Return list of ancestors, closest last."""
        mro = list(self.__class__.__mro__)
        mro.reverse()        
        ancestors = mro[1:-1] # neither proper class nor object is counted as ancestor
        return ancestors
        
    def set_attributes(self, provided_dict):
        for (member_name, value) in provided_dict.items():
            self.__setattr__(member_name, value)
        
    def init(self):
        pass
    
    
    



    

class Named(LeanDescendant):
    """Common base class that provides some more or less helpfull
    general purpose methods. If there is no reason against, we make 
    objects always of from this type.
    """
     
    def init(self, 
                name       = None,
                verbosity  = 1,  
                ):    
        """
        @param name: string to name the object
        @param verbosity: the verbosity level of the object, 0 meaning dump
            If a message is to print, the level of the message is compared to 
            verbosity of the object and only passed if the latter is not lower
            the message level. 
            
        @todo: the message/verbosity handling should be redone applying 
            the python logging framework 
        """ 
        
        self.last_tick = {}
        
        
    def get_name(self):
        if self.has('name') and self.name:
            return self.name
        else:
            return self.__class__.__name__
       
    

    def send(self,  *messages, **parameters):
        """
        Passes logging messages to stdout.
        
        Since we a *-argument we cannot name the parameters in the signature.
        The parameters are:
        level: from which verbosity of self should the message be passed
        linebreaks, lb: number of linebreaks to append to the message
        """
        
        ## digest parameters    
        if parameters.has_key('level'):
            message_level = parameters['level']
        else:
            message_level = 1
            
        if   parameters.has_key('linebreaks'):
            n_linebreaks = parameters['linebreaks']
        elif parameters.has_key('lb'):
            n_linebreaks = parameters['lb']
        else:
            n_linebreaks = 1


        ## go
        if self.verbosity >= message_level:
        
            for part in messages:
                print part,
            
            if n_linebreaks > 0:
                print '\n' * (n_linebreaks-1)
                        
            sys.stdout.flush()
        
                

    def clock(self, clock=None,restart=False):
        
        if self.last_tick.has_key(clock):
            new_tick              = time.time()
            delta                 = new_tick - self.last_tick[clock]
            self.last_tick[clock] = new_tick
        else:
            self.last_tick[clock] = time.time()
            delta = 0.
        
        return delta



    def send_clock(self, clock=None, message=None):
        
        
        delta = self.clock(clock)
                
        if message:
            text = message
        else:
            if clock is None:
                clock_name = self.get_name()
            else:
                clock_name = str(clock)
            text = 'clock ' + clock_name + ':'
        
        self.send(text, seconds2time(delta))
        


    def warnings(self,error_text=''):
        self.warning('Warning: ', error_text)
 
    def bugger(self,error_text=''):
        self.send('Bugger: ', error_text)
        raise Exception(error_text) 
    
    def abstract_method_bugger(self,error_text=''):
        self.send('Abstract Method Bugger: ', error_text)
        raise Exception(error_text)

    def not_implemented_bugger(self,error_text=''):
        self.send('Not yet implemented. ', error_text)
        raise Exception(error_text)    


    def shell(self,command_fields, level=2, send_errors=False):
        """
        @param command_fields: parts of command as taken by subprocess.Popen
        @param level: verbosity level
        
        call a single command command_fields[0] with args
        command_fields[1:]. Cares for quotation, and escapes.
        
        @todo: verbosity level management should be integrated 
            python logging framework
        """
        
    
        p = Popen(command_fields  , shell=False, stdout=PIPE, stderr=PIPE )
        
        self.send('shell call:', command_fields,  level=level)        
        stdout_content,  stderr_content = p.communicate()
        
        self.send(stdout_content, level=level)  
            
        if stderr_content:
            out = stdout_content + '\n' +  stderr_content
            if send_errors:
                self.send(stderr_content)
        else:
            out = stdout_content 
        
        self.send(out,  level=level) 
        return out


    def has(self, attribute_name):
        try:
            self.__getattribute__(str(attribute_name))
        except AttributeError:
            return False
        else:
            return True
        
    def flag(self, attribute_name):
        return self.has(attribute_name) and self.__getattribute__(str(attribute_name))


    def added_members(self):
        
        dummy = self.__class__()
        
        members_dummy = dummy.__dict__.keys()
        members       = self.__dict__.keys()
        
        extra = list(set(members).difference(set(members_dummy)))
        
    
        return dict([ (k,self.__dict__[k]) for k in extra ])
    
    

class Applicable(Named):
    """Wrapper around functions"""
    
    def init(self, 
             apply = None,
            ):    
        pass
          
    def apply(self):
        self.abstract_method_bugger()




class Singleton(object):
    
    def __eq__(self, x):
        
        return type(self) == type(x)


class EmptyArg(Singleton):
    pass


class Boxed(Named):
    """Class Box: Container for objects that need a wrapper class"""
    def init(self, content):
        self.content = content
        pass


class Object(object):
    """Can serve as container with .-access to compound."""
    pass
 


class Empty(object):
    pass



# for testing
class R(Named):
    def init(self, a=234, b=33):
        print "RR0", self    
        self.a += 50
        print "RRR", self.a,b

class V(R):
    def init(self,y=2323):  
        print self  
        print "VVV", y
        print "VV2",  self.y, self.b

class D(V):
    def init(self, g, r=11, a=5):
        print self
        print "DDD"
        print self.__dict__
        

def I(indexable):
    return range(len(indexable))



 
if __name__ == '__main__':
    
    
    A = Named()
    A.clock()
    time.sleep(1)
    A.send_clock()
    
    
