"""Smoothed Objectives for Binary Classification."""

from numpy import array, ones, shape, concatenate, dot, sum, linalg,\
     inf, sqrt, reshape, size, zeros, where, diag
from scipy import rand, stats, linalg
     
from tb.baseclass import Named
#from tb.plotting import FunctionPlot
from tb.scipytools import tp, row, col, a2sign

from tb.mixedtools import transpose_list, I
from ml.objective import one_dim_cut_factory
from openopt import NLP
from pw_toolbox.odict import OrderedDict


### definition of example classes

class ExampleClass(Named):
    """class defining a subset of examples."""

    def is_member(self,x,y):
        self.abstract_method_bugger()

    def select_examples(self,X,Y):
        
        selection = [ l for l in I(X) if self.is_member(X[l], Y[l]) ]
                    
        return X[selection], col(Y[selection])
    


class Positives(ExampleClass):
    def is_member(self,x,y):
        return y == +1
    
class Negatives(ExampleClass):
    def is_member(self,x,y):
        return y == -1


### regularization

class Regulizer(Named):
    def init(self,
             smoothed = True,
             sigma    = 1.,
             apply_B  = False,
             M    = None,
             n    = None):
        
        if apply_B:
            assert M
            self.B  = M.B
            self.K_ = M.K_
            #self.S  = M.S
        else:
            if M:
                self.B  = diag(ones(M.q))
                self.K_ = diag(ones(M.q))
                #self.S  = M.q
                
        pass
    
    def R_total(self,v):
        
        R = self.sR_total if self.smoothed else self.uR 
        out = R(v).ravel()
        assert size(out) == 1

        return out

    def R_dv_total(self,v):
        return self.sR_dv_total(v)

    def R_dv_2_total(self,v):
        return self.sR_dv_2_total(v)
      
    def get_w(self,v):
        w  = v[:-1,0]
        if self.apply_B:
            w = dot(self.B,w)
        return w
    
    def get_R_dw(self,v):
        w  = v[:-1,0]
        if self.apply_B:
            R_dw = row( 2* dot(self.K_, w))
        else:
            R_dw = row( 2*  w)
        return R_dw
    
class QuadraticR(Regulizer):
    
 
    def uR(self,v):        
        w_ = self.get_w(v)        
        return dot(tp(w_),w_)
                    
    ### smoothed
    def sR_total(self,v):
        w_ = self.get_w(v)  
        q  = len(w_)      
        out =  ( dot(tp(w_),w_) + q * self.sigma**2 ).ravel()
        #print "\nR = %f = %f + %f * %f" % (out,dot(tp(w_),w_), self.S , self.sigma**2) 
        return out
     
    def sR_dv_total(self,v):
        
        R_dw = self.get_R_dw(v) 
        R_db = array([0], ndmin=2)
        out = concatenate((R_dw,R_db), axis=1).ravel()
        return out
    
    def sR_dv_2_total(self,v):
        w = v[:-1,0]
        n = len(w)
        R_dw_2 = self.K_
        
        R_db_2_1 = zeros((1,n))
        R_db_2_2 = zeros((n+1,1))
        
        R_dv_2   = concatenate((R_dw_2, R_db_2_1), axis=0)
        R_dv_2   = concatenate((R_dv_2, R_db_2_2), axis=1)
        
        assert shape(R_dv_2) == (n+1, n+1)
     
        return R_dv_2
    




### losses

class LossPart(Named):
    "Loss part: base function for parts of a loss function"""

    
    def h_(self, t, a_, b_):
        """The base function of the compound."""
        self.abstract_method_bugger()

    def H(self):
        """The indefinite integral of the base function h_."""
        self.abstract_method_bugger()



class LinearLP(LossPart):
    """linear loss component"""

    def h_(self, m_):
        return  m_
                    
    def H(self, t, a_, b_):

        return  - a_ *  p(t)  + b_ * F(t) 

    def H_dt(self, t, a_, b_):

        return  - a_ *  p_(t) - a_ * F(t) +  b_ * p(t)
    
    def H_dt_2(self, t, a_, b_):

        return  - a_ *  p__(t)   - 2. * a_ * p(t) +  b_ * p_(t)
    


class QuadraticLP(LossPart):
    """linear loss component"""

    def h_(self, m_):
        return  m_**2
                    
    def H(self, t, a_, b_):
        return   a_**2 * p_(t)  - a_ * b_ * p(t) +  (a_**2 + b_**2) * F(t) 
                 
    def H_dt(self, t, a_, b_):

        return   a_**2 * p__(t)  + ( 2 * a_**2 + b_**2) * p(t) +\
               - a_ * b_ * p_(t) + ( 2 *  b_  * -a_ )  * F(t)  
    
    def H_dt_2(self, t, a_, b_):

        return    a_**2 * p___(t)  - 4 * a_ * b_ * p(t) + ( 3 * a_**2 + b_**2 ) * p_(t)\
               -  a_ * b_ * p__(t) + ( 2 * a_**2 )  * F(t)  
      
    




class Loss(Named):
    """Margin based loss function that is split up to an indefinite number
    of loss parts."""
    def init( self,
              LP     = [ LinearLP, LinearLP ],  
              QR     = [ (-inf, 1.), (1., inf)  ],  
              QR_bar = [ (0., 1.), (0., 1.) ], 
              CD     = [ (-1., 0.), (0., 0.) ],
              q0_p  = None,
              rk_p  = None,
              smoothed = True,
              sigma    = 1.,
              dv_2     = False,
              ):
        pass
        """Constructor of a loss function.
        @param LP: list of loss parts classes 
        @param QR: list of intervals [[q_0,r_0],..,[q_k,r_k]] on which the corresponding 
            (by index) loss parts are defined
        @param QR_bar: list of base intervals: the loss part takes the shape of 
            the base function on this interval.
        @param q0_p: the first an the last in interval QR may be improper,
             q_0 = -inf, r_k = +inf, such we cannot derive the linear transforms
            (a_0,b_0), (a_k,b_k) in these cases. Therefore, we give a finite 
            proxy values q0_p and rk_p to define the shape on border intervals
            by defining it on [q0_p,r_0] and [q_k,rk_p] respectively. 
            By default q0_p = r_0 - 1 and rk_p = q_k + 1.
        @param rk_p: see q0_p,
        @param smoothed: wether the loss filtered through Gaussian smoothing by default
        @param sigma: standard deviation of Gaussian used for smoothing
        
        The intervals in QR may overlap and do not need to cover the real axis.
        Overlapping losses means that losses are added up and uncovered areas
        correspond to zero loss.    
        """
        
        self.n_loss_parts = self.k = k = len(LP)
        
        self.q,self.r           = transpose_list(QR)
        self.q                  = [ float(q) for q in self.q ]
        assert len(self.q) == k
                
        self.c, self.d          = transpose_list(CD)
        assert len(self.c) == k
        

        self.q_bar, self.r_bar  = transpose_list(QR_bar)

                                    
        self.a = []
        self.b = []
        for i in range(k):
            q,r         = self.q[i], self.r[i]
            q_bar,r_bar = self.q_bar[i], self.r_bar[i]

            ## if interval [q,r] is improper,
            ## set self.q_bar[0]/self.r[k] to improper values as well
            ## but override q/r with proxy values for computing a and b
            if q == -inf:
                assert i is 0
                if q0_p is None:
                    q0_p = r - 1.
                q = q0_p 
                self.q_bar[0] = -inf
                 
            if r == +inf:
                assert i is k-1
                if rk_p is None:
                    rk_p = q + 1.
                r = rk_p 
                self.r_bar[-1] = +inf
                
            ## now compute a and b
            a           = (q_bar - r_bar)/(q - r)
            b           = - a * q + q_bar 
            assert not a == 0
            self.a += [ a ]
            self.b += [ b ]
        

        self.a = array(self.a, ndmin=2)
        self.b = array(self.b, ndmin=2)
        

    
    def prepare(self,X,Y):
        
        ## initialize loss parts
        for i in I(self.LP):
            if type(self.LP[i]) is type(Named):
                self.LP[i] = self.LP[i]() 
            
        self.n_examples = len(X)
        self.X      = X
        self.Y      = Y
        self.X_norm = reshape(sqrt(sum(X**2,axis=1)), (len(X),1))
        self.X_     = self.X / ( self.X_norm * self.sigma )
        self.a_     = self.a * self.Y * self.X_norm * self.sigma   
        X   = self.X_
        l,d = shape(X)
        X_1 = reshape(X, (l,d,1))
        X_2 = reshape(X, (l,1,d))
        if self.dv_2:
            self.XX  = X_1 * X_2    
            assert shape(self.XX) == (l,d,d)
        assert shape(self.a_) == (len(X), size(self.a) )
        
        self.Z = col(self.Y) * self.X
        

       
    def E_total(self,v, smoothed=None):
        
        smoothed = self.smoothed if smoothed is None else smoothed 
        if smoothed:
            return sum(sum(self.sE(v,A='H', B=F)))
        else:
            return sum(sum(self.uE(v)))

    def E_dv_total(self,v):
        return self.sE_dv_total(v)
        
    def E_dv_2_total(self,v):
        return self.sE_dv_2_total(v)


    def sE_dv_total(self,v):
        
        E   = self.sE(v, A='H_dt', B=p)
        out = sum( col(sum(E, axis=1)) * - self.X_ , axis=0 )
        
        return out    
    
    
    def sE_dv_2_total(self,v):
        X   = self.X_
        l,d = shape(X)
        E   = self.sE(v, A='H_dt_2', B=p_)    
        out = sum( reshape(sum(E, axis=1), (l,1,1)) * self.XX, axis=0 )        
        assert shape(out) == (d,d)

        return out    
    
        
    
    def uE(self,v):
        
        m   = self.Y * dot(self.X,v)    
        m_  = self.a * m + self.b
        hm_ = zeros((self.n_examples,self.k))
        
        for i in range(self.k):
            hm_[:,i] = self.LP[i].h_(m_[:,i])
       
        out =  ( m >= self.q ) * ( m < self.r ) * (self.c * hm_ + self.d) 
        
        assert shape(out) == (self.n_examples,self.k) 
        return out 
    
        
    def sE(self,v, A='H', B=None):
        v = col(v)
        m        = self.Y * dot(self.X,v)
        a_       = self.a_
        b_       = m * self.a + self.b
        q_breve  = (self.q_bar - b_) / a_    
        r_breve  = (self.r_bar - b_) / a_    
        
        Hq_breve = zeros((self.n_examples,self.k))
        Hr_breve = zeros((self.n_examples,self.k))
        for i in range(self.k):
            H = self.LP[i].__getattribute__(A)
            q_breve_i = col(q_breve[:,i])
            r_breve_i = col(r_breve[:,i])
            a__i      = col(a_[:,i])
            b__i      = col(b_[:,i])
            Hq_breve[:,i] = H( q_breve_i, a__i, b__i ).ravel()
            Hr_breve[:,i] = H( r_breve_i, a__i, b__i ).ravel() 

        out  =  a2sign(a_) * ( \
                row(self.c) * ( Hr_breve   - Hq_breve) + \
                row(self.d) * ( B(r_breve) - B(q_breve) ) )
    
        
        assert shape(out) == shape(b_)
        return out 



class HingeLoss(Loss):
    
    def init( self,
              LP     = [ LinearLP,   ],  
              QR     = [ (-inf, 1.)  ],  
              QR_bar = [ (0, 1.),    ], 
              CD     = [ (-1., +1.)  ],
              
              ):
        pass
    
class HingeLoss2(Loss):
    
    def init( self,
              LP     = [ LinearLP,   ],  
              QR     = [ (-inf, 1.)  ],  
              QR_bar = [ (0, 1.),    ], 
              CD     = [ (-2., +2.)  ],
              
              ):
        pass
    

class QuadraticHingeLoss(Loss):
    
    def init( self,
              LP     = [ QuadraticLP,  LinearLP ],  
              QR     = [ (-inf, 1.), (1., +inf)  ],  
              QR_bar = [ (-1,  0.),    (0., 1) ], 
              CD     = [ (1., 0.),   (0., 0.) ],
              ):
        pass


class SteepQuadraticLoss(Loss):
    
    def init( self,
              LP     = [ QuadraticLP, ],  
              QR     = [ (-inf,  1.),  ],  
              QR_bar = [ (-1.5,   -0.5),   ], 
              CD     = [ (1.,  -1./4),    ],
              ):
        pass

class SteepQuadraticLoss2(Loss):
    
    def init( self,
              LP     = [ QuadraticLP, ],  
              QR     = [ (-inf,  1.),  ],  
              QR_bar = [ (-2,   -1.),   ], 
              CD     = [ (1./3,  -1./3),    ],
              ):
        pass
    
class SteepQuadraticLoss3(Loss):
    
    def init( self,
              LP     = [ QuadraticLP,  LinearLP ],  
              QR     = [ (-inf,  1./2),    (1./2, 1)  ],  
              QR_bar = [ (-3./2,   -1./2), (0, 1./2)  ], 
              CD     = [ (1.,     +1./4),  (-1, 1./2)  ],
              ):
        pass



class BoundedHingeLoss(Loss):
    """Hinge loss bounded from above by 2."""
    def init( self,
              LP     = [ LinearLP, LinearLP, LinearLP ],  
              QR     = [ (-inf, -1), (-1, 1.), (1., inf)  ],    
              QR_bar = [ (0, 1),     (0, 1),   (0, 1) ], 
              CD     = [ (0., 2.),   (-2, 2.), (0., 0.) ],
              ):
        pass
    

class SeparationLossA(Loss):
    
    def init( self,
              LP     = [   LinearLP, LinearLP ],  
              QR     = [  (0.5,  1), (1, inf)  ],  
              QR_bar = [  (0,  1),     (0,  1) ], 
              CD     = [  (-1, 0),     (0., -1.), ],
              ):
        pass
    
class SeparationLossA0(Loss):
    
    def init( self,
              LP     = [   LinearLP, LinearLP ],  
              QR     = [  (0,  1),     (1, inf)  ],  
              QR_bar = [  (0,  1),     (0,  1) ], 
              CD     = [  (-1, 0),     (0., -1.), ],
              ):
        pass
    
class SeparationLossA2(Loss):
    
    def init( self,
              LP     = [   LinearLP, LinearLP ],  
              QR     = [  (0.5,  1.),    (1., inf)  ],  
              QR_bar = [  (0,    0.5),     (0,  1) ], 
              CD     = [  (-1.,  0),     (0., -1/2.), ],
              ):
        pass


class SeparationLossA2p(Loss):
    
    def init( self,
              LP     = [   LinearLP, LinearLP ],  
              QR     = [  (-inf,0.5), (0.5,  1.), ],  
              QR_bar = [  (0,   1),  (0,  1) ], 
              CD     = [  (0.,  0.5),  (-0.5, 0.5), ],
              ):
        pass

class SeparationLossA2pv(Loss):
    
    def init( self,
              LP     = [   LinearLP, LinearLP , LinearLP],  
              QR     = [  (-inf,0.5), (0.5,  1.), (1., inf)],  
              QR_bar = [  (0,   1),  (0,  1),(0,  1) ], 
              CD     = [  (0.,  0.5),  (-0.5, 0.5), (2.**-16,0.) ],
              ):
        pass

class SeparationLossA2pv1(Loss):
    
    def init( self,
              LP     = [   LinearLP, LinearLP ],  
              QR     = [  (-inf,0), (0,  1.),],  
              QR_bar = [  (0,   1), (0,  1), ], 
              CD     = [  (0.,  1), (-1, 1),  ],
              ):
        pass


    
class SeparationLossA3(Loss):
    
    def init( self,
              LP     = [   LinearLP, LinearLP ],  
              QR     = [  (0.5,  1.5),    (1.5, inf)  ],  
              QR_bar = [  (0,    1),      (0,  1) ], 
              CD     = [  (-1.,  0),      (0., -1.), ],
              ):
        pass



class SeparationLossB(Loss):
    
    def init( self,
              LP     = [  LinearLP, ],  
              QR     = [  (0, inf), ],  
              QR_bar = [  (0, 1),   ], 
              CD     = [  (0., -1.) ],
              ):
        pass

  
class SeparationLossC2(Loss):
    
    def init( self,
              LP     = [ LinearLP, LinearLP, LinearLP ],  
              QR     = [ (-1,   0),  (0,   1),     (1., inf)  ],  
              QR_bar = [ (0,    1),  (0,   1),     (0, 1) ], 
              CD     = [ (1./2,  0),  (-1, +1./2), (0,-1./2) ],
              ):
        pass


class SeparationLossD(Loss):
    
    def init( self,
              LP     = [ LinearLP, LinearLP, LinearLP, LinearLP, LinearLP ],  
              QR     = [ (-2, -1), (-1,0), (0, 1),     (1.,2), (2,3)  ],  
              QR_bar = [ (0,  1),  (0, 1), (0, 1),     (0, 1), (0,1)  ], 
              CD     = [ (1., 0),  (0, 1), (-2., +1.), (0,-1), (1,-1) ],
              ):
        pass


class ACompactLoss(Loss):
    
    def init( self,
              LP     = [ LinearLP,  LinearLP ],  
              QR     = [ (-1, 0.), (0,  1)  ],  
              QR_bar = [ ( 0, 1.), (0., 1) ], 
              CD     = [ (1., 0.), (-1., +1.) ],
              
              ):
        pass
    
    
class AnotherCompactLoss(Loss):
    
    def init( self,
              LP     = [ LinearLP,   LinearLP, LinearLP ],  
              QR     = [ (-6,-5), (-5,0), (0,  1)  ],  
              QR_bar = [ ( 0, 1.),  (0., 1), (0., 1) ], 
              CD     = [ (1., 0.), (0,+1), (-1., +1.) ],
              
              ):
        pass




class RMLPO(Named):
    """Regularized Multiple Loss Parts Objective.
    """
    def init(self,
             regulizers = [ QuadraticR ],
             classes    = [ Negatives, Positives ],
             losses     = [ HingeLoss, HingeLoss ],
             weights    = [ 1 , 3, 1  ],
             X          = array([], ndmin=2),
             Y          = array([], ndmin=2),
             sigma      = 1.,
             smoothed   = True,
             M          = None,
             apply_BR   = False,
             n          = None,
             dv_2       = False):
        
        ## prepare losses
        pass
        self.instanciate()
        
    def instanciate(self):
            
        self.n_classes    = len(self.classes)
        self.n_regulizers = len(self.regulizers)
        assert len(self.losses) == self.n_classes
        assert len(self.weights) == self.n_regulizers + self.n_classes 

        for j in range(self.n_regulizers):
            
            if type(self.regulizers[j]) is type:
                self.regulizers[j] = self.regulizers[j](sigma=self.sigma,
                                                        smoothed=self.smoothed,
                                                        M=self.M,
                                                        apply_B=self.apply_BR,
                                                        ) 
        
        self.class_sizes = []
        for j in range(self.n_classes):
            if type(self.classes[j]) is type:
                self.classes[j] = self.classes[j]() 
            if type(self.losses[j]) is type:
                self.losses[j]  = self.losses[j](sigma=self.sigma,
                                                 smoothed=self.smoothed,
                                                 dv_2 = self.dv_2)
            
            Xj,Yj           = self.classes[j].select_examples(self.X,self.Y)
            self.class_sizes += [len(Xj)] 
            self.losses[j].prepare(Xj,Yj)
        
        self.class_sizes = array(self.class_sizes)

    def set_smoothing(self, b):
        self.smoothed = b
        for loss in self.losses:
            loss.smoothed = b
        for r in self.regulizers:
            r.smoothed = b
    
                

    def Q_generic(self,v, what=''):
                    
        v   = col(v) # any objective-part that is called from has to do this
        
        RF  =  [ regulizer.__getattribute__('R_' + what + 'total') 
                 for regulizer in self.regulizers ]
        
        n_r = len(RF)
        assert n_r > 0
        
        r   = array( [ R(v) for R in RF ] )
        
        weights_r = array( self.weights[:n_r]) * 1./(v.size -1)
        
        weights_r  = reshape(weights_r , [ n_r ] + [ 1 ] * len(shape(r[0])) ) 
        
        EF   = [ loss.__getattribute__('E_' + what + 'total')
                 for loss in self.losses ]
        
        e    = array( [ E(v) for E in EF ] )
        
        
        
        n_e = len(EF)
        assert n_e > 0
        
        weights_e  = array( self.weights[n_r:]) * 1./self.class_sizes

        weights_e  = reshape( weights_e, [ n_e ] + [ 1 ] * len(shape(e[0])) ) 
        

            
        out =   sum(weights_r * r, axis=0) +\
                sum(weights_e * e, axis=0)       
        
        if what=='':    
            #print "Q = %f = %f * %f  +  %f * %f + %f * %f " % \
            #     ( out, weights_r[0],r[0],  weights_e[0],e[0] ,  weights_e[1],e[1])
            pass
        
        assert shape(out) == shape(r[0])
        
        return out 
        
    def Q(self,v):
        
        return self.Q_generic(v, what='')

    def Q_dv(self,v):
        
        return self.Q_generic(v, what='dv_')

    def Q_dv_2(self,v):
        
        return self.Q_generic(v, what='dv_2_')


class SVML1(RMLPO):
    
    def init(self,
             losses     = [ HingeLoss, HingeLoss ],
             ):
        pass
        


class SVML2(RMLPO):
    
    def init(self,
             losses     = [ QuadraticHingeLoss, QuadraticHingeLoss ],
             ):
        pass
     

class SVMB(RMLPO):
    
    def init(self,
             losses     = [ BoundedHingeLoss, BoundedHingeLoss ],
             ):
        pass
    
    
class SVMF(RMLPO):
    
    def init(self,
             losses     = [ HingeLoss, BoundedHingeLoss ],
             ):
        pass
    
class SVMG(RMLPO):
    
    def init(self,
             losses     = [ HingeLoss, AnotherCompactLoss ],
             ):
        pass

class SecoObjectiveA0(RMLPO):
    
    def init(self,
             losses     = [ HingeLoss2, SeparationLossA ],
             ):
        pass

class SecoObjectiveA1(RMLPO):
    
    def init(self,
             losses     = [ HingeLoss, SeparationLossA2 ],
             ):
        pass
    
class SecoObjectiveA2(RMLPO):
    
    def init(self,
             losses     = [ HingeLoss, SeparationLossA2 ],
             ):
        pass
    
class SecoObjectiveA2p(RMLPO):
    
    def init(self,
             losses     = [ HingeLoss, SeparationLossA2p ],
             ):
        pass
    
class SecoObjectiveA2pv(RMLPO):
    
    def init(self,
             losses     = [ HingeLoss, SeparationLossA2pv ],
             ):
        pass
    
class SecoObjectiveA2pv1(RMLPO):
    
    def init(self,
             losses     = [ HingeLoss, SeparationLossA2pv ],
             ):
        pass


class SecoObjectiveA3(RMLPO):
    
    def init(self,
             losses     = [ HingeLoss, SeparationLossA3 ],
             ):
        pass

class SecoObjectiveA4(RMLPO):
    
    def init(self,
             losses     = [ SteepQuadraticLoss3, SeparationLossA3 ],
             ):
        pass
       
       
class SecoObjectiveB0(RMLPO):
    
    def init(self,
             losses     = [ HingeLoss, SeparationLossB ],
             ):
        pass


class SecoObjectiveC(RMLPO):
    
    def init(self,
             losses     = [ SteepQuadraticLoss, SeparationLossC2 ],
             ):
        pass

class SecoObjectiveC2(RMLPO):
    
    def init(self,
             losses     = [ SteepQuadraticLoss, SeparationLossA2 ],
             ):
        pass

class SecoObjectiveC3(RMLPO):
    
    def init(self,
             losses     = [ SteepQuadraticLoss, SeparationLossA3 ],
             ):
        pass


class SecoObjectiveD(RMLPO):
    
    def init(self,
             losses     = [ HingeLoss, SeparationLossD ],
             ):
        pass


class SecoObjectiveE(RMLPO):
    
    def init(self,
             losses     = [ SteepQuadraticLoss, SeparationLossA2 ],
             ):
        pass

class SecoObjectiveF(RMLPO):
    
    def init(self,
             losses     = [ SteepQuadraticLoss, SeparationLossA0 ],
             ):
        pass

class SecoObjectiveF0(RMLPO):
    
    def init(self,
             losses     = [ HingeLoss, SeparationLossA0 ],
             ):
        pass

class SecoObjectiveF2(RMLPO):
    
    def init(self,
             losses     = [ HingeLoss2, SeparationLossA0 ],
             ):
        pass
    
class SecoObjectiveF3(RMLPO):
    
    def init(self,
             losses     = [ SteepQuadraticLoss3, SeparationLossA0 ],
             ):
        pass


class Testing(Named):
    
    def init(self,objective=None):
        pass

    def check_gradient(self, 
                       loss_class, n=20, d = 12,  sigma=0.2,
                       n_checks=4):
            
        
        X = ( rand(n,d) - 0.5) * 20
        Y = a2sign(rand(n,1) -0.5 )
        
        
        #print Y
        #Y =  ones((n,1))
        
        sob = RMLPO(X=X,Y=Y,losses = [ loss_class, loss_class ], 
                   sigma=sigma,
                   weights=[0.1,1,1],
                   smoothed=True
                   )   
        y = 0
        
        e      = sob.losses[y].E_total
        e_dv   = sob.losses[y].sE_dv_total
        e_dv_2 = sob.losses[y].sE_dv_2_total
        
        
        def f(x):
            return sob.Q(x)
        
        def f_dv(x):
            return sob.Q_dv(x)

        def f1_dv(x):
            return sob.Q_dv(x)[1]

        def f1_dv_2(x):
            return sob.Q_dv_2(x)[1]

        
        for i in range(n_checks):
            v_0 = self.random_vector(d, scale=3.)
            NLP( f1_dv, v_0 , df = f1_dv_2).checkdf()    
    
    

    def random_vector(self,d, scale=1):
        return (rand(d)-0.5)*scale*rand(d)
      
    def one_dim_cut_factory(self,f, v_0 , v_1):
        """for f:R^n -> B, a closure cut:R -> B
        with cut(r) = f(v_0 + r v_1_) is produced 
        where v_1_ = v_1/||v_1||"""
        
        v_0 = col(v_0)
        v_1 = col(v_1)
        
        assert shape(v_0) == shape(v_1)
        
        def cut(r):
            return f( v_0 + r * (1./ linalg.norm(v_1)) * v_1)
        
        return cut
        
    def directional_derivative_factory(self, grad, v_0, v_1):
        
        grad_cut = one_dim_cut_factory(grad, v_0, v_1)
    
        def dd(r):
            return dot( grad_cut(r), col(v_1)/linalg.norm(v_1) )
        
        return dd

    def plot_e(self, n=32,d=32, sigma=1./8):
        
        X = ( rand(n,d) - 0.5) * 2
        Y = a2sign(rand(n,1) - 0.5 )
            
        v_0 = col(self.random_vector(d, scale=2))
        v_1 = col(self.random_vector(d, scale=1))
       
    
        sigmas = [1./4, 1./16 ]
        
        sobs = OrderedDict()
        sfs  = OrderedDict()
        
        
        for sigma in sigmas:
            sobs[sigma] = SecoObjectiveA2pv1(X=X,Y=Y, sigma=sigma, weights=[0.02,1,1])
            sobs[sigma].set_smoothing(True)
            es    = self.one_dim_cut_factory(sobs[sigma].Q, v_0 , v_1)
            es.__name__ = 'smoothed, '+  r'$\sigma=%s$' % str(sigma) 
            
            sfs[sigma] = es
            
            
        
        uob = SecoObjectiveA2pv1(X=X,Y=Y, sigma=sigma,weights=[0.02,1,1],)
        uob.set_smoothing(False)
        
        eu    = self.one_dim_cut_factory(uob.Q, v_0 , v_1)
        eu.__name__ = 'unsmoothed'
    
        g = FunctionPlot( 
                          functions =  [  eu ] + sfs.values(),  
                          colors    = ['b'] + ['g'] * len(sfs),
                          x_interval = [-3.,3.],
                          n_points=200,
                          tick_argmax = False,
                          tick_max = False,
                          title = 'random cut through objective on random examples',#$%d random point with %d dimension' % ( n,d),
                          #% (str([ loss.get_name() for loss in sob.losses]),n),
                          x_label = r'$\lambda$',
                          y_label = r'$Q(v)$',
                          show_legend=True,
                          linewidth=2,
                          x_ticks  = [-1,0,1],
                         )
        g.go()
        

 
    
    def plot_m_e(self,loss_class,sigma=0.2):
    
        #X = array([ [1.], [1.],[1.],[1] ] , ndmin=2 )
        #Y = array([ [1.], [1.],[1.],[1] ] , ndmin=2 )
        X = array([ [1.]] , ndmin=2 )
        Y = array([ [1.]] , ndmin=2 )     
        #print "AAA", shapes(X,Y)

        
        sob = loss_class(sigma=sigma)
        sob.prepare(X,Y)
                
        v = col(array([1.] * len(X[0])))
        
        def smoothed_loss(m):
            #return sum([ loss.E_total(v) for loss in sob.losses ] ).ravel()
            sob.smoothed = True
            #print shapes(m,v, m*v, sob.E_total(m*v))
            return sob.E_total(m*v)
        
        smoothed_loss.__name__ = "smoothed_loss" + r' $\sigma=%1.1f$' % sigma
            
        
        def loss(m):
            sob.smoothed = False
            return sob.E_total(m*v)
        
        
                
        g = FunctionPlot( show_legend=0,#True,
                          functions =  [ loss ],#, smoothed_loss ],               
                          #x_lim=[0.2,1.2]
                          x_interval=[-1.5,1.5],
                          y_lim=[-1.5,1.5],

                          x_label = r'$m$',
                          y_label = r'$L(m)$',
                          title=sob.get_name(),
                          tick_argmax = False,
                          tick_max = 0,
                          tick_min = 0,
                          x_ticks  = [-1,0,1],
                          y_ticks  = [-1,0,1],
                          #text = r'$\sigma=%f1.1' % sigma,
                          n_points=300,
                          linewidth=2,
                        )
        g.go()
        



## Gaussian distribution functions 
SND = stats.norm(0,1)

p   = SND.pdf

F   = SND.cdf


def p_(v) :
    return where(abs(v) < inf, - v * p(v), 0.)

def p__(v): 
    return where(abs(v) < inf, - ( p(v) + v * p_(v) ), 0.)

def p___(v): 
    return where(abs(v) < inf, - ( 2 * p_(v) + v * p__(v) ), 0.)
 

 
if __name__ == '__main__' and 1:
    
    from tb.plotting import FunctionPlot
    import pylab
    OT = Testing()
    
    #OT.check_gradient(QuadraticHingeLoss)
        
    #OT.plot_m_e(SeparationLossA, sigma=0.5)
    
    #OT.plot_m_e(SeparationLossA2pv1, sigma=0.2)
    #OT.plot_m_e(HingeLoss, sigma=0.2)
    for i in range(10):
        pylab.figure(i)
    
        OT.plot_e( )
    
    pylab.show()
        
        
    if 0:
        a = 2.
        b = 1.
        
        def g(t):
            return (a * t + b ) *  p(t) 
            
        def h(t):
            return -a * p_(t) - a * F(t) + b * p(t)
        
        
        def H_(t):
            return (a*t + b ) * p(t)
        
        
        g = FunctionPlot( show_legend=True,
                              functions =  [ H_ ],               
                              #x_lim=[0.2,1.2]
                              #x_interval=[-5,5]
                              tick_argmax = False,
                              n_points=100
                            )
    
        g.go()
    
    #pylab.show()

            
        
                    
        
       
        

        
        
    
    